/*
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.joni.constants.internal;

public enum TokenType {
      EOT,            /* end of token */
      RAW_BYTE,
      CHAR,
      STRING,
      CODE_POINT,
      ANYCHAR,
      CHAR_TYPE,
      BACKREF,
      CALL,
      ANCHOR,
      OP_REPEAT,
      INTERVAL,
      ANYCHAR_ANYTIME,  /* SQL '%' == .* */
      ALT,
      SUBEXP_OPEN,
      SUBEXP_CLOSE,
      CC_OPEN,
      QUOTE_OPEN,
      CHAR_PROPERTY,    /* \p{...}, \P{...} */
      LINEBREAK,
      EXTENDED_GRAPHEME_CLUSTER,
      KEEP,
      /* in cc */
      CC_CLOSE,
      CC_RANGE,
      POSIX_BRACKET_OPEN,
      CC_AND,             /* && */
      CC_CC_OPEN          /* [ */
}
